/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "SrcInfo.h"
#include "Vector.h"
#include "Map.h"

using namespace std;
using namespace FD;

class CvfdhFeatureHoughLine;
   
DECLARE_NODE(CvfdhFeatureHoughLine)
/*Node
 *
 * @name CvfdhFeatureHoughLine
 * @category FDOPENCV_HARK:Feature
 * @description CvfdhFeatureHoughLine 
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @input_name SRCIN
 * @input_type Vector<ObjectRef>
 * @input_description Source location. A type of an element of the vector is SrcInfo.
 *
 * @output_name FEATOUT
 * @output_type Map<int,ObjectRef>
 * @output_description Output features. The ObjectRef should be Vector<float>.
 *
 * @parameter_name ALGORITHM
 * @parameter_type string
 * @parameter_value STANDARD
 * @parameter_list STANDARD:PROBABLISTIC
 * @parameter_description
 *
 END*/
   
   
class CvfdhFeatureHoughLine : public BufferedNode {

  int m_imageInID;
  int m_srcInID;
  int m_featOutID;
  string algorithm;
      
public:
  CvfdhFeatureHoughLine(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params), m_srcInID(-1)
  {
    m_imageInID = addInput("IMAGEIN");
    m_featOutID = addOutput("FEATOUT");
    algorithm = object_cast<String>(parameters.get("ALGORITHM"));
    inOrder = true;
  }      

  virtual int translateInput(string inputName) {
    if (inputName == "SRCIN") {
      m_srcInID = addInput(inputName);
      return m_srcInID;
    }
    else if (inputName == "IMAGEIN") {
      for (unsigned int i = 0; i < inputs.size(); i++) {
	if (inputs[i].name == inputName) {
	  return i;
	}
      }
    }
    else {throw new NodeException(this, inputName+ " is not supported.", __FILE__, __LINE__);}
  }
      
  void calculate(int output_id, int count, Buffer &out)
  {

    RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
    (*(outputs[m_featOutID].buffer))[count] = outputMapPtr;

    RCPtr<Map<int, ObjectRef> > image_map(new Map<int, ObjectRef>);

    // Making the Map<id, IplImage*> for processing
    IplImage* img_in_tmp;
    ObjectRef in = getInput(m_imageInID, count);
    if (typeid(*in) == typeid(CvfdImage)) {
      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	return;
      }
      img_in_tmp = cvCloneImage(imagePtr->getImage());
      CvfdImage* image= new CvfdImage(img_in_tmp);  
      (*image_map)[0] = image;
    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {
      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	int store_flag = 1;
	if(m_srcInID != -1){
	  RCPtr<Vector<ObjectRef> > sources_ptr = getInput(m_srcInID, count);
	  const Vector<ObjectRef>& sources = *sources_ptr;
	  store_flag = 0;
	  for (int srci = 0; srci < sources.size(); srci++) {
	    RCPtr<SrcInfo> src = sources[srci];
	    if(src->id == it->first) store_flag = 1;
	  }
	}
	if(store_flag == 0) continue;
	img_in_tmp = cvCloneImage(imagePtr.getImage());
	CvfdImage* image= new CvfdImage(img_in_tmp);  
	(*image_map)[it->first] = image;
      }
    }

    // Main routine
    Map<int, ObjectRef>::iterator img_it = image_map->begin();
    for (; img_it != image_map->end(); ++img_it) {
      IplImage* img_in;      
      CvfdImage& imagePtr = object_cast<CvfdImage>(img_it->second);
      img_in = cvCloneImage(imagePtr.getImage());
    
      RCPtr<Vector<float> > map_contents(new Vector<float>);
    
      // ---------------------------------------------------------------
      // main routine : start  

      if(img_in->nChannels > 1){
	IplImage *gray;
	gray = cvCreateImage( cvGetSize(img_in),IPL_DEPTH_8U,1);
	cvCvtColor(img_in, gray, CV_BGR2GRAY);
	img_in = cvCloneImage (gray);
	cvReleaseImage (&gray);
      }    

      IplImage* dst = cvCreateImage( cvGetSize(img_in), 8, 1 );
      IplImage* color_dst = cvCreateImage( cvGetSize(img_in), 8, 3 );
      CvMemStorage* storage = cvCreateMemStorage(0);
      CvSeq* lines = 0;
      int i;
      cvCanny( img_in, dst, 50, 200, 3);
      cvCvtColor( dst, color_dst, CV_GRAY2BGR );

      if(algorithm == "STANDARD"){
	lines = cvHoughLines2( dst, storage, CV_HOUGH_STANDARD, 1, CV_PI/180, 100, 0, 0 );
	for( i = 0; i < MIN(lines->total,100); i++ ){
	  float* line = (float*)cvGetSeqElem(lines,i);
	  float rho = line[0];
	  float theta = line[1];
	  CvPoint pt1, pt2;
	  double a = cos(theta), b = sin(theta);
	  double x0 = a*rho, y0 = b*rho;
	  pt1.x = cvRound(x0 + 1000*(-b));
	  pt1.y = cvRound(y0 + 1000*(a));
	  pt2.x = cvRound(x0 - 1000*(-b));
	  pt2.y = cvRound(y0 - 1000*(a));
	  // cvLine( color_dst, pt1, pt2, CV_RGB(255,0,0), 3, 8 );
	  (*map_contents).push_back((float)line[0]);
	  (*map_contents).push_back((float)line[1]);
	}
      }

      if(algorithm == "PROBABLISTIC"){
	lines = cvHoughLines2( dst, storage, CV_HOUGH_PROBABILISTIC, 1, CV_PI/180, 50, 50, 10 );
	for( i = 0; i < lines->total; i++ ){
	  CvPoint* line = (CvPoint*)cvGetSeqElem(lines,i);
	  // cvLine( color_dst, line[0], line[1], CV_RGB(255,0,0), 3, 8 );
	  float* line2 = (float*)cvGetSeqElem(lines,i);
	  (*map_contents).push_back((float)line2[0]);
	  (*map_contents).push_back((float)line2[1]);
	} 
      }

      cvReleaseImage (&img_in);
      cvReleaseImage (&dst);
      cvReleaseImage (&color_dst);

    }

    cvReleaseImage(&img_in_tmp);

  }
   
  NO_ORDER_NODE_SPEEDUP(CvfdhFeatureHoughLine)
};
