/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include <gdk/gdk.h>

using namespace std;
using namespace FD;

class CvfdShowImage;
   
DECLARE_NODE(CvfdShowImage)
/*Node
 *
 * @name CvfdShowImage
 * @category FDOPENCV_ORG:Image
 * @description View an image
 *
 * @input_name IMAGEIN
 * @input_description Image to visualize
 * @input_type CvfdImage
 *
 * @parameter_name WINDOW_TITLE
 * @parameter_type string
 * @parameter_value BaseImage
 * @parameter_description Window title.
 * 
 * @parameter_name FLAGS
 * @parameter_type int
 * @parameter_value 1
 * @parameter_description If it is set to 1, window size is automatically adjusted to fit the displayed image, while user can not change the window size manually.
 *
 * @output_name IMAGEOUT
 * @output_description Same as input image 
 * @output_type CvfdImage
 *
 END*/
   
   
class CvfdShowImage : public BufferedNode {
      
  int m_imageInID;
  int m_imageOutID;
  std::string m_windowName;
  int m_flags;
      
      
public:
  CvfdShowImage(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    //add inputs
    m_imageInID = addInput("IMAGEIN");
    //add outputs
    m_imageOutID = addOutput("IMAGEOUT");         
    //Initialize parameters         
    RCPtr<String> windowNamePtr = parameters.get("WINDOW_TITLE");
    m_flags = dereference_cast<int>(parameters.get("FLAGS"));
    //Create the windows
    m_windowName = *windowNamePtr;
    gdk_threads_enter();
    cvNamedWindow(m_windowName.c_str(), m_flags );
    gdk_threads_leave();
  }
      
  ~CvfdShowImage()
  {
    gdk_threads_enter();
    cvDestroyWindow( m_windowName.c_str() );
    gdk_threads_leave();  
  }
      
  void calculate(int output_id, int count, Buffer &out)
  {
    //Read the inputs
    RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);
         
    //Handle
    gdk_threads_enter();
    int status = cvGetErrMode();
    cvSetErrMode( CV_ErrModeSilent );
    __BEGIN__; 
    OPENCV_CALL(cvShowImage(m_windowName.c_str(), imagePtr->getImage()));   
    __END__;
    cvSetErrMode( status );
    if( cvGetErrStatus() != CV_StsOk  )
      {
	throw new GeneralException("OPENCV - Error to show the image: " +  CCHAR(cvErrorStr( cvGetErrStatus() )),__FILE__,__LINE__);
      }          
    gdk_threads_leave();         
    out[count] = imagePtr;
  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdShowImage)
};
